<?php

/**
 * @file
 * Functions to build designed transaction forms.
 */

use Drupal\mcapi\Mcapi;
use Drupal\mcapi\Entity\Type;
use Drupal\mcapi\Entity\WalletInterface;
use Drupal\system\Entity\Menu;
use Drupal\user\Entity\Role;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Form\FormState;
use Drupal\Core\Url;
use Drupal\Core\Link;
use Drupal\Core\Render\Element\PathElement;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\Markup;
use Drupal\Core\Entity\Entity\EntityFormMode;

/**
 * Theme preprocessor.
 *
 * Break the form up into 'fields', 'hidden', and actions, where the fields
 * appear rendered in the twig template
 */
function template_preprocess_first_party_transaction_form(&$vars) {
  // Add the buttons at the end if they weren't in the Twig template already.
  $form = &$vars['form'];

  $renderer = \Drupal::service('renderer');
  $vars['actions'] = $renderer->render($form['actions']);

  $hidden = '';
  foreach (Element::Children($form) as $field_name) {
    $element = $form[$field_name];
    if (isset($element['#type']) and ($element['#type'] == 'hidden' or $element['#type'] == 'token')) {
      $hidden .= $renderer->render($element);
    }
    else {
      if($field_name === 'worth') {
        unset($element['widget']['#title']);
        unset($element['widget']['#description']);
      }
      else {
        mcapi_strip_wrappers_recursive($element);
      }
      $tokens[$field_name] = $renderer->render($element);
    }
  }
  $vars['hidden'] = Markup::create($hidden);
  $vars['fields'] = [
    '#type' => 'inline_template',
    '#template' => _filter_autop($form['#twig_template']),
    '#context' => $tokens,
  ];
}

/**
 * Helper function.
 */
function mcapi_strip_wrappers_recursive(array &$element) : void {
  // Dont' strip any composite elements
  $composite_element_types = ['checkbox', 'checkboxes', 'radios', 'worths_form'];
  if (isset($element['#type']) && in_array($element['#type'], $composite_element_types)) {
    return;
  }
  unset($element['#theme_wrappers'], $element['#title']);
  $element['#description_display'] = 'hidden';
  foreach (Element::Children($element) as $delta) {
    mcapi_strip_wrappers_recursive($element[$delta]);
  }
}

/**
 * Alter the form EntityFormDisplay form.
 *
 */
function mcapi_forms_display_decorate(&$form, FormState $form_state, $settings) {
  $w = 0;
  ksort($form['fields']);
  // show any extra fields and non-Basefields
  $hide = ['type', 'created', 'creator'];
  // prevent moving fields to the hidden region.
  foreach (Element::children($form['fields']) as $field_name) {
    unset($form['fields'][$field_name]['region']['#options']['hidden']);
    $form['fields'][$field_name]['region']['#default_value'] = in_array($field_name, $hide) ? 'hidden' : 'content';
  }
  // hide the hidden region with hidden fields in it
  unset($form['fields']['#regions']['hidden']);


  $form['title'] = [
    '#title' => t('Title of the form'),
    '#type' => 'textfield',
    '#default_value' => $settings['title']??'',
    '#size' => 40,
    '#maxlength' => 80,
    '#required' => TRUE,
    '#weight' => -10,
  ];

  $bill_options = ['' => t("All roles with 'create bill transaction' permission")];
  $credit_options = ['' => t("All roles with 'create credit transaction' permission")];
  foreach (Role::loadMultiple() as $rid => $role) {
    if ($role->hasPermission('create bill transaction')) {
      $bill_options[$rid] = $role->label();
    }
    if ($role->hasPermission('create credit transaction')) {
      $credit_options[$rid] = $role->label();
    }
  }
  $form['incoming'] = [
    '#title' => t('Direction of the transaction'),
    '#description' => t('In relation to the user.'),
    '#type' => 'radios',
    '#options' => [
      '1' => t('Incoming (bill)'),
      '0' => t('Outgoing, (credit)')
    ],
    '#default_value' => $settings['incoming']??NULL,
    '#required' => TRUE,
    '#weight' => -9,
  ];

  $form['bill_role'] = [
    '#title' => t('Restricted to role'),
    '#description' => t("These roles all have 'create bill transaction' permission "),
    '#type' => 'radios',
    '#options' => $bill_options,
    '#default_value' => $settings['role'] ?? '',
    '#multiple' => TRUE,
    '#disabled' => count($bill_options) < 2,
    '#states' => [
      'visible' => [
        ':input[name="incoming"]' => ['value' => 1],
      ]
    ],
    '#weight' => -8,
    'user1' => [
      '#default_value' => TRUE,
      '#disabled' => TRUE
    ]
  ];

  $form['credit_role'] = [
    '#title' => t('Restricted to role'),
    '#description' => t("These roles all have 'create credit transaction' permission "),
    '#type' => 'radios',
    '#options' => $credit_options,
    '#default_value' => $settings['role'] ?? '',
    '#multiple' => TRUE,
    '#disabled' => count($credit_options) < 2,
    '#states' => [
      'visible' => [
        ':input[name="incoming"]' => ['value' => 0],
      ]
    ],
    '#weight' => -8,
    'user1' => [
      '#default_value' => TRUE,
      '#disabled' => TRUE
    ]
  ];


  $form['routing'] = [
    '#title' => t('Routing & links'),
    '#type' => 'details',
    '#collapsible' => TRUE,
    '#weight' => -5
  ];
  $form['routing']['path'] = [
    '#title' => t('Path'),
    '#description' => t('The url path at which this form will appear. Must begin with / and be unique.'),
    '#type' => 'textfield',
    '#default_value' => @$settings['path'] ?: '/',
    '#placeholder' => '/path/to/form',
    '#convert_path' => PathElement::CONVERT_URL,
    '#element_validate' => ['mcapi_forms_unique_path'],
    '#required' => TRUE,
    '#field_prefix' => \Drupal::service('router.request_context')->getCompleteBaseUrl(),
  ];

  $form['routing']['menu'] = [
    '#title' => t('Menu link'),
    '#type' => 'fieldset',
    '#tree' => 0,
    '#weight' => 5,
    'menu_title' => [
      '#title' => t('Link title'),
      '#type' => 'textfield',
      '#default_value' => isset($settings['menu_title']) ? $settings['menu_title'] : '',
      '#weight' => 1,
    ],
    'menu_weight' => [
      '#title' => t('Weight'),
      '#type' => 'weight',
      '#default_value' => isset($settings['menu_weight']) ? $settings['menu_weight'] : 0,
      '#description' => t('In the menu, the heavier links will sink and the lighter links will be positioned nearer the top.'),
      '#weight' => 2,
    ],
  ];
  $form['routing']['wallet_link_title'] = [
    '#title' => t('Link title to show on wallet'),
    '#description' => t('Link to this form from the wallet. Optionally use token [mcapi_wallet:name]'),
    '#type' => 'textfield',
    '#weight' => 5,
    '#default_value' => @$settings['wallet_link_title'],
  ];
  if (!\Drupal::moduleHandler()->moduleExists('menu_ui')) {
    \Drupal::service('module_installer')->install(['menu_ui']);
  }
  $custom_menus = [];
  foreach (Menu::loadMultiple() as $menu_parent => $menu) {
    $custom_menus[$menu_parent] = $menu->label();
  }
  $form['routing']['menu']['menu_parent'] = \Drupal::service('menu.parent_form_selector')->parentSelectElement(
    isset($settings['menu_parent']) ? $settings['menu_parent'] : 'main:',
    'menu_link_content:7299677d-ac7b-4748-a2b8-d22cac2c9b62',
    menu_ui_get_menus()
  );
  $form['routing']['menu']['menu_parent']['#empty_value'] = '';

  $form['transaction_type'] = [
    '#title' => t('Transaction type'),
    '#description' => t('This also determines the starting state of the transaction.'),
    '#type' => 'select',
    '#options' => Mcapi::entityOptionsList(Type::loadMultiple()),
    '#default_value' => @$settings['transaction_type'],
    '#weight' => -8,
    '#required' => TRUE,
  ];

  list($required, $optional) = mcapi_forms_required_fields();
  // Remove invisible fields
  $invisible = ['xid', 'uuid', 'serial', 'parent', 'creator', 'type', 'state', 'changed', 'created'];

  $required = t(
    'Required fields: @fields.',
    ['@fields' => ' {{ ' . implode(' }}, {{ ', array_diff($required, $invisible)) . ' }}']
  );
  $optional = t(
    'Optional fields: @fields.',
    ['@fields' => ' {{ ' . implode(' }}, {{ ', array_diff($optional, $invisible)) . ' }}']
  );

  $form['experience'] = [
    '#title' => t('User experience'),
    '#type' => 'fieldset',
    '#open' => TRUE,
    'experience_twig' => [
      '#title' => t('Main form @Twig', ['@Twig' => 'Twig']),
      '#description' => implode(' ', [
        t('Use the following @Twig tokens with HTML & css to design your payment form. Linebreaks will be replaced automatically.', ['@Twig' => 'Twig']),
        $required .' '.$optional . ' ' .
        Link::fromTextAndUrl(
          t('What is @Twig?', ['@Twig' => 'Twig']),
          Url::fromUri('http://twig.sensiolabs.org/doc/templates.html')
        )->toString(),
      ]),
      '#type' => 'textarea',
      '#rows' => 6,
      '#default_value' => str_replace(array('\r\n', '\n', '\r'), "\n", @$settings['experience_twig']),
      '#element_validate' => ['validate_twig_template'],
      '#weight' => 1,
      '#required' => TRUE,
    ],
    'experience_button' => [
      '#title' => t('Button label'),
      '#description' => t("The text to appear on the 'save' button, or the absolute url of an image."),
      '#type' => 'textfield',
      '#default_value' => @$settings['experience_button'],
      '#required' => TRUE,
      '#weight' => 2,
    ],
//    'experience_preview' => [
//      '#title' => t('Preview mode'),
//      '#type' => 'radios',
//      '#options' => [
//        TransactionActionBase::CONFIRM_NORMAL => t('Basic - Go to a fresh page'),
//        TransactionActionBase::CONFIRM_AJAX => t('Ajax - Replace the form'),
//        TransactionActionBase::CONFIRM_MODAL => t('Modal - Confirm in a dialogue box'),
//      ],
//      '#default_value' => @$settings['experience_preview'],
//      '#weight' => 3,
//      '#required' => TRUE,
//    ],
//    '#weight' => $w++,
  ];

  $form['#suffix'] = t(
    "N.B The confirmation page is configured separately, at %link",
    [
      '%link' => Link::fromTextAndUrl(
      'admin/accounting/workflow/save',
      Url::fromRoute('entity.action.edit_form', ['action' => 'transaction_save'])
      )->toString(),
    ]
  );
  array_unshift($form['actions']['submit']['#submit'], 'mcapi_forms_display_submit');
  if (!empty($settings)) {
    // Redirect only if this is not the first saving.
    $mode = $form_state->getFormObject()->getEntity()->getMode();
    $form_state->setRedirectUrl(Url::fromRoute("mcapi.1stparty.$mode"));
  }
}

/**
 * Check the required transaction form elements are either preset OR in Twig.
 */
function validate_twig_template(array $element, $form_state) {
  list($required) = mcapi_forms_required_fields();
  $invisible = ['xid', 'uuid', 'serial', 'parent', 'creator', 'created', 'type', 'state', 'changed'];
  unset($required['creator']);
  // Find the required field and ensure they are all here.
  foreach (array_diff($required, $invisible) as $fieldname) {
    if (!mcapi_forms_has_twig_token($element['#value'], $fieldname)) {
      $form_state->setError(
        $element,
        t(
          '@fieldname MUST appear in the @title template.',
          [
            '@fieldname' => '{{ '.$fieldname.' }}',
            '@title' => $element['#title'],
          ]
        )
      );
    }
  }
  $field_definitions = \Drupal::service('entity_field.manager')
    ->getFieldDefinitions('mcapi_transaction', 'mcapi_transaction');
  // Find the fields in the template and check they are all enabled.
  $enabled_fields = $form_state->getFormObject()->getEntity()->getComponents();
  $tokens = extractTwigTokens($element['#value']);
  if (count($tokens) <> count(array_unique($tokens))) {
    $form_state->setError($element, t('Each twig token can only appear once in the form.'));
    return;
  }
  foreach ($tokens as $fname) {
    if (!isset($field_definitions[$fname])) {
      $form_state->setError($element, t('Invalid Twig token @token', ['@token' => '{{ '.$fname.' }}']));
      return;
    }
    elseif (!isset($enabled_fields[$fname])) {
      $form_state->setError(
        $element,
        t(
          "'@fieldname' field must be enabled and configured to be included in the form template.",
          ['@fieldname' => $field_definitions[$fname]->getLabel()]
        )
      );
      return;
    }
  }
  // This assumes neither payer nor payee is duplicated
  preg_match("/\{\{ ?(?:payee|payer) ?\}\}/", $element['#value'], $matches);
  if (count($matches) < 1 or count($matches) > 2) {
    $form_state->setError($element, t('Payer and/or payee field must be included.'));
  }
}

/**
 * Form element validation ensuring that the given path is unique.
 */
function mcapi_forms_unique_path(&$element, $form_state) {
  $value = $form_state->getValue('path');
  if (!substr($value, 0, 1) == '/') {
    $form_state->setError($element, t('Paths must begin with a /'));
  }
  $url = Url::fromUserInput($form_state->getValue('path'));
  if ($url->isRouted()) {
    if (substr($url->getRouteName(), 15) != $form_state->getFormObject()->getEntity()->get('mode')) {
      \Drupal::messenger()->addWarning(t('N.B. This has overriden an existing path!'));
    }
  }
}

/**
 * Form submit callback.
 * Save the entity_display_form settings in the corresponding entity_display_mode
 */
function mcapi_forms_display_submit($form, $form_state) {
  $mode = $form_state->getFormObject()->getEntity()->getMode();
  \Drupal::entityTypeManager()->clearCachedDefinitions();
  $entity_display_mode = EntityFormMode::load('mcapi_transaction.'.$mode);
  $fields = [
    'title',
    'incoming',
    'wallet_link_title',
    'transaction_type',
    'path',
    'menu_title',
    'menu_weight',
    'menu_parent',
    'experience_twig',
    'experience_button',
  ];

  foreach ($fields as $settingname) {
    $entity_display_mode->setThirdpartySetting(
      'mcapi_forms',
      $settingname,
      $form_state->getValue($settingname)
    );
  }
  $roles_field = $form_state->getValue('incoming') ? 'bill_role' : 'credit_role';
  $entity_display_mode->setThirdpartySetting(
    'mcapi_forms',
    'role',
    $form_state->getValue($roles_field)
  );
  $entity_display_mode->save();
}

/**
 * Get the transaction fields, required, and optional
 *
 * @return array[]
 *   Containing 2 arrays, one of required field names, the other optional.
 */
function mcapi_forms_required_fields() {
  $fields = \Drupal::service('entity_field.manager')->getFieldDefinitions('mcapi_transaction', 'mcapi_transaction');
  foreach ($fields as $field_id => $field) {
    if ($field->isrequired()) {
      $required[] = $field_id;
    }
    else {
      $optional[] = $field_id;
    }
  }
  return [$required, $optional];
}

/**
 * Migration helper.
 * Get only the translated field names used in the template.
 *
 * @param string $template
 * @param string $direction
 *
 * @return array
 *   The field names used in the templated, mapped to the modern names.
 */
function mcapi_forms_migrate_template_field_map($template, $direction) {
  //Get the tokens used in the old template
  preg_match_all('/\[mcapiform:([a-z_]+)\]/', $template, $matches);
  $old_tokens = $matches[1];
  unset($old_tokens[array_search('direction', $old_tokens)]);
  return mcapi_forms_migrate_field_map($old_tokens, $direction);
}

/**
 * Get the map of all old field names to new
 * @param array $defaults
 * @param string $direction
 *
 * @return array
 *   The old and new fieldnames
 */
function mcapi_forms_migrate_field_map($defaults, $direction) {
  $mapping = [
   "offers_wants_categories" => "category",
   "transaction_description" => "description"
  ] + array_combine($defaults, $defaults);
  if ($direction == 'outgoing') {
    $mapping['secondperson'] = 'payee';
  }
  else {
    $mapping['secondperson'] = 'payer';
  }
  return $mapping;
}

/**
 *
 * @param string $template
 * @param string $tok
 * @return bool
 */
function mcapi_forms_has_twig_token(string $template, string $tok) : bool {
  return preg_match("/\{\{ ?$tok ?\}\}/", $template);
}

/**
 * Create links to the payment forms, prepopulated if possible.
 *
 * Iterate through the transaction form displays and create links to ones using
 * the 'my_wallet' widget.
 *
 * @param WalletInterface $wallet
 *
 * @return array[]
 *   Items each containing a Url and string value
 */
function _mcapi_forms_quick_links(WalletInterface $wallet = NULL) : array {
  $info = [];
  $ids = \Drupal::entityTypeManager()->getStorage('entity_form_display')->getQuery()->accessCheck(FALSE)
    ->condition('id', 'mcapi_transaction.mcapi_transaction.default', '<>')
    ->condition('bundle', 'mcapi_transaction')
    ->execute();
  foreach (EntityFormDisplay::loadMultiple($ids) as $entity_form_display) {
    $entity_form_mode = EntityFormMode::load('mcapi_transaction.'.$entity_form_display->getMode());
    if (!$entity_form_mode) {
      $id = $entity_form_display->id();
      \Drupal::logger('Community Accounting')->error(
        "Transaction form mode does not exist for EntityFormDisplay '@id' with mode @mode",
        ['@id' => $id, '@mode' => $entity_form_display->getMode()]
      );
      continue;
    }
    $incoming = $entity_form_mode->getThirdPartySetting('mcapi_forms', 'incoming');
    if (is_null($incoming)) {
      continue; // It means the form direction hasn't been set, since the setting was created.
    }
    $options = [];
    if ($incoming == 1) {
      $options['query'] = ['payer' => $wallet->id()];
    }
    elseif ($incoming === 0) {
      $options['query'] = ['payee' => $wallet->id()];
    }
    else {
      //continue;
    }
    if ($wallet) {
      $title = str_replace('[mcapi_wallet:name]', $wallet->label(), $entity_form_mode->getThirdPartySetting('mcapi_forms', 'wallet_link_title'));
    }
    else {
      $title = $entity_form_mode->getThirdPartySetting('mcapi_forms', 'title');
    }
    $path = $entity_form_mode->getThirdPartySetting('mcapi_forms', 'path');
    if (!substr($path, 0, 1)== '/') {
      $path = '/'.$path;
    }
    $info[$entity_form_mode->id()] = [
      'url' => Url::fromUserInput($path, $options),
      'title' => $title
    ];
  }

  return $info;
}
