<?php

namespace Drupal\mcapi\Entity\ViewBuilder;

use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Form\FormState;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityViewBuilder;
use Drupal\Core\Render\Markup;

/**
 * Visualisations of transactions per currency.
 */
class CurrencyViewBuilder extends EntityViewBuilder {

  /**
   * {@inheritdoc}
   */
  protected function getBuildDefaults(EntityInterface $currency, $view_mode) {
    $build = parent::getBuildDefaults($currency, $view_mode);
    $build['#title'] = $this->t('Trade stats for @currency', ['@currency' => $currency->name]);
    $build['description'] = [
      '#markup' => Markup::create('<p>'.t('Description'). ': '. $currency->description.'</p>'),
      '#weight' => -50
    ];

    if ($view_mode == 'summary') {
      return $build;
    }

    unset($build['#theme']);
    $transactionStorage  = \Drupal::entityTypeManager()->getStorage('mcapi_transaction');

    $form_state = (new FormState())->set('currency', $currency);
    $build['filter'] = [
      '#title' => $this->t('Filter transactions...'),
      '#type' => 'details',
      // @todo inject this
      'form' => \Drupal::formBuilder()->buildForm('Drupal\mcapi\Form\TransactionStatsFilterForm', $form_state),
      '#weight' => -1,
      '#open' => isset($_SESSION['transaction_stats_filter'])
    ];

    $build['#cache_contexts'] = ['accounting'];

    if ($earliest = $currency->firstUsed()) {
      // $since needs to be set to the second previous to the first transaction;
      $conditions['since'] = $earliest - 1;
    }
    $conditions =isset($_SESSION['transaction_stats_filter']) ?  array_filter($_SESSION['transaction_stats_filter']) : [];

    $balances_asc = $trades = $volumes = $incomes = $spending = [];

    $rows = $transactionStorage->ledgerStateByWallet($currency->id(), $conditions);
    foreach ($rows as $row) {
      $balances_asc[$row->wid] = (int)$row->balance;
      $trades[$row->wid] = $row->trades;
      $volumes[$row->wid] = (int)$row->income + (int)$row->expenditure;
      $incomes[$row->wid] = (int)$row->income;
      $spending[$row->wid] = (int)$row->expenditure;
    }

    asort($balances_asc);
    // Get the bottom wallets.
    for ($i = 0; $i < min(10, count($balances_asc)/2); $i++) {
      $wid = key($balances_asc);
      $wallet = Wallet::load($wid);
      if (!$wallet) {
        throw new \Exception("No wallet $wid in \Drupal\mcapi\Entity\ViewBuilder\CurrencyViewBuilder");
      }
      $quant = current($balances_asc);
      $bottoms[] = [
        'raw' => abs($quant),
        'link' => $wallet->toUrl(),
        'worth' => $currency->format($quant, CurrencyInterface::DISPLAY_NORMAL),
        'name' => $wallet->label(),
      ];
      next($balances_asc);
    }
    // Get the top wallets.
    $balances_desc = array_reverse($balances_asc, TRUE);
    for ($i = 0; $i < min(10, count($balances_asc)/2); $i++) {
      $wid = key($balances_desc);
      $wallet = Wallet::load($wid);
      if (!$wallet) {
        throw new \Exception("No wallet $wid in \Drupal\mcapi\Entity\ViewBuilder\CurrencyViewBuilder");
      }
      $quant = current($balances_desc);
      $tops[] = [
        'raw' => $quant,
        'link' => $wallet->toUrl(),
        'worth' => $currency->format($quant, CurrencyInterface::DISPLAY_NORMAL),
        'name' => $wallet->label(),
      ];
      next($balances_desc);
    }

    $build['balances_chart'] = [
      '#theme' => 'mcapi_extreme_balances',
      '#curr_id' => $currency->id(),
      '#depth' => 10,
      '#class' => ['extreme-balances'],
      '#largest'  => max(max($balances_desc), abs(min($balances_desc))),
      '#tops' => $tops,
      '#bottoms' => $bottoms,
      '#weight' => -1,
    ];

    $build['trades_per_user_wallet'] = [
      '#title' => t('Trades per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'trades_per_user_wallet_' . $currency->id(),
      '#curr_id' => $currency->id(),
      '#users_only' => TRUE,
      '#format_vals' => FALSE,
      '#data' => $trades,
      '#top' => 5,
    ];
    $build['volume_per_user_wallet'] = [
      '#title' => t('Volume per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'volume_per_user_wallet_' . $currency->id(),
      '#curr_id' => $currency->id(),
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $volumes,
      '#top' => 5,
    ];
    $build['income_per_user_wallet'] = [
      '#title' => t('Income per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'income_per_user_wallet_' . $currency->id(),
      '#curr_id' => $currency->id(),
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $incomes,
      '#top' => 5,
    ];
    $build['spending_per_user_wallet'] = [
      '#title' => t('Expenditure per wallet'),
      '#type' => 'mcapi_ordered_wallets',
      '#id' => 'spending_per_user_wallet_' . $currency->id(),
      '#curr_id' => $currency->id(),
      '#users_only' => TRUE,
      '#format_vals' => TRUE,
      '#data' => $spending,
      '#top' => 5,
    ];

    //@todo support time periods
    list(, , $period, $period_name) = $this->periodQueryParams($conditions);
    list($dates, $trades, $volumes, $wallets) = $transactionStorage->historyPeriodic($currency->id(), $period, $conditions);

    $params = [
      '@units' => $currency->label(),
      '@span' => strtolower($period_name), // this should be already translated, somehow.
    ];
    $build['trades_history'] = [
      '#theme' => 'mcapi_timeline',
      '#title' => t('Numbers of @units trades per @span', $params),
      '#points' => array_combine($dates, $trades),
      '#width' => 800,
      '#height' => 200,
    ];
    $build['participation_history'] = [
      '#theme' => 'mcapi_timeline',
      '#title' => t('Numbers of traders per @span', $params),
      '#points' => array_combine($dates, $wallets),
      '#width' => 800,
      '#height' => 200,
    ];
    $build['volume_history'] = [
      '#theme' => 'mcapi_timeline',
      '#title' => t('Volumes of @units traded per @span', $params),
      '#points' => array_combine($dates, $volumes),
      '#currency' => $currency,
      '#width' => 800,
      '#height' => 200,
    ];

    $build['#attached']['library'][] = 'mcapi/mcapi.currency';
    return $build;
  }

  /**
   * Work out the start, end and frequency to poll the db.
   *
   * Less than a week, invalid.
   * Less than 100 days.
   * Less than 100 weeks, weekly.
   * Less than 100 months, monthly.
   * Otherwise annual.
   *
   * @param array $conditions
   * @return array
   *   $from, $to, the time period, and the timeperiod translated
   */
  public function periodQueryParams(array $conditions) : array {
    $since = isset($conditions['since']) ? $conditions['since'] : 0;
    $until = isset($conditions['until']) ? $conditions['until'] : REQUEST_TIME;
    $span = $until - $since;
    $day = 86400;
    // $year = 31560192;
    if ($span < $day * 7) {
      // This won't produce very good results.
      $from = date('d-m-Y', $since);
      $to = date('d-m-Y', $until + $day);
      $groupby = 'Day';
    }
    // 100 days.
    elseif ($span < $day * 100) {
      $from = date('d-m-Y', $since);
      $to = date('d-m-Y', $until + $day);
      $groupby_translated = t('Day');
      $groupby = 'day';
    }
    // 100 weeks.
    elseif ($span < $day * 700) {
      // Get the beginning of the week.
      $from = date('Y', $since) . 'W' . date('W', $since);
      // The beginning of the week after the end.
      $to = date('Y', $since) . 'W' . (date('W', $since) + 1);
      $groupby_translated = t('Week');
      $groupby = 'week';
    }
    // 100 months.
    elseif ($span < $day * 3000) {
      $from = date('Y', $since);
      // The beginning of the week after the end.
      $to = date('Y', $since) + 1;
      $groupby_translated = t('Month');
      $groupby = 'month';
    }
    else {
      $from  = date('Y', $since);
      $to = date('Y', $until) + 1;
      $groupby_translated = t('Year');
      $groupby = 'year';
    }
    return [strtotime($from), strtotime($to), $groupby, $groupby_translated];
  }

}
