<?php

namespace Drupal\mcapi\Form;

use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\user\Entity\User;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\TempStore\PrivateTempStoreFactory;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Routing\RouteMatchInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Session\AccountProxy;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Form for 3rdParty transactions, and base form for 1st party transactions.
 */
class TransactionForm extends ContentEntityForm {

  const DIR_INCOMING = 1;
  const DIR_OUTGOING = 0;

  /**
   * The tempStore service.
   *
   * @var Drupal\Core\TempStore\PrivateTempStore
   */
  protected $tempStore;

  /**
   * The request object.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * The current user.
   *
   * @var \Drupal\user\Entity\User
   */
  protected $currentUser;

  /**
   * The logger channel.
   *
   * @var Drupal\Core\Logger\LoggerChannel;
   */
  protected $logger;

  /**
   *
   * @param $entity_repository
   * @param $entity_type_bundle_info
   * @param $time
   * @param PrivateTempStoreFactory $tempstore
   * @param Request $current_request
   * @param AccountProxy $current_user
   * @param LoggerChannelFactory $logger_channel
   */
  public function __construct($entity_repository, $entity_type_bundle_info, $time, PrivateTempStoreFactory $tempstoreFactory, Request $current_request, AccountProxy $current_user, LoggerChannel $logger_channel) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->tempStore = $tempstoreFactory->get('TransactionForm');
    $this->request = $current_request;
    $this->currentUser = User::load($current_user->id());
    $this->logger = $logger_channel;
  }

  /**
   * {@inheritdoc}
   *
   * @todo update to entity_type.manager
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('tempstore.private'),
      $container->get('request_stack')->getCurrentRequest(),
      $container->get('current_user'),
      $container->get('logger.channel.mcapi')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $display = $this->getFormDisplay($form_state);
    $mode = $display->getOriginalMode();
    // Forms designer module has its own modes and must set the transaction direction itself.
    if ($mode == 'bill') {
      $form_state->set('transaction_direction', static::DIR_INCOMING);// make this a constant.
    }
    elseif($mode == 'credit') {
      $form_state->set('transaction_direction', static::DIR_OUTGOING);
    }

    $display->buildForm($this->entity, $form, $form_state);
    return $form;
  }

  /**
   * {@inheritdoc}
   *
   * @note we are overriding here because this form is neither for saving nor
   * deleting and because previewing is not optional.
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    return [
      'submit' => [
        '#type' => 'submit',
        '#value' => $this->t('Preview'),
        // N.B. NOT ::save
        '#submit' => ['::submitForm'],
      ],
    ];
  }

  public function validateForm(array &$form, FormStateInterface $form_state) {
    $entity = parent::validateForm($form, $form_state);
    // The only place to add children in Transaction::validate(), but in the
    // submit phase the transaction is rebuilt without validation. So we store
    // the children in form_state and add them back in the submit phase.
    // If is was done with an entityBuilder it wouldn only work for forms.
    $form_state->set('children', $entity->children);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);
    // Re-add the children created in the validation phase.
    $this->entity->children = $form_state->get('children');
    $this->tempStore->set('mcapi_transaction', $this->entity);
    $form_state->setRedirect(
      'entity.mcapi_transaction.save',
      ['mcapi_transaction' => 0]
    );
  }

  /**
   * {@inheritdoc}
   *
   * @todo test creating a transaction with and without specifying the creator
   */
  public function buildEntity(array $form, FormStateInterface $form_state) {
    $entity = parent::buildEntity($form, $form_state);

    // If the wallets aren't populated, use the first wallet of the current user.
    if ($entity->payee->isEmpty()) {
      $entity->payee->target_id = WalletStorage::firstWalletOfEntity($this->currentUser);
    }
    elseif ($entity->payer->isEmpty()) {
      $entity->payer->target_id = WalletStorage::firstWalletOfEntity($this->currentUser);
    }

    if (!$form_state->getValue('created')) {
      $form_state->setValue('created', REQUEST_TIME);
    }
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function getEntityFromRouteMatch(RouteMatchInterface $route_match, $entity_type_id) {
    // Assumes that any query params will be for the transaction.
    if ($props = $this->request->query->all()) {
      $entity = Transaction::create($props);
    }
    else {
      $entity = parent::getEntityFromRouteMatch($route_match, $entity_type_id);
    }
    return $entity;
  }

}
